module Main exposing (main)

import FunkyFunktions exposing (..)
import Stickers exposing (Sticker, StickerContent(..), viewSticker)

import Array exposing (Array, get, set)
import Array.Extra as Array
import Browser
import Browser.Dom exposing (focus, getViewport, Viewport)
import Browser.Navigation exposing (Key)
import Browser.Events exposing (onMouseMove, onResize)
import Html exposing (Attribute, button, div, h3, input, Html, section, text)
import Html.Attributes exposing (class, id, style, value)
import Html.Attributes.Extra as Attributes exposing (attributeIf)
import Html.Extra exposing (viewMaybe)
import Html.Events exposing (onBlur, onClick, onInput, onMouseEnter)
import Json.Decode as D
import List exposing (map, singleton)
import Maybe.Extra exposing (isJust)
import String exposing (fromInt, isEmpty)
import String.Extra exposing (isBlank)
import Svg exposing (Svg, svg)
import Svg.Attributes exposing (height, viewBox, width)
import Svg.Events exposing (on)
import Task
import Url exposing (Url)

main = Browser.application
  { init = init
  , view = view
  , update = update
  , onUrlChange = constant Noop
  , onUrlRequest = constant Noop
  , subscriptions = subscriptions
  }

type Msg
  = Noop
  | Click Int Int Int Int
  | SetTileText Int Int String
  | AddTile Int
  | TileDeselected Int
  | HeldOverNewTile Int Int
  | NewMousePos Int Int

type alias Model =
  { columns: Array Column
  , heldTile: Maybe HeldTile
  , mousePos: (Int, Int)
  }

type alias Column =
  { name: String
  , tiles: Array Tile
  }

type alias Tile =
  { text: String
  , stickers: List Sticker
  }

type alias HeldTile =
  { hoveredColumnIndex: Int
  , hoveredTileIndex: Int
  , tile: Tile
  }

blankTile = { text = "", stickers = [] }

init : () -> Url -> Key -> (Model, Cmd Msg)
init flags url browserKey =
  { columns =
    [ { name = "Into"
      , tiles = Array.empty
      }
    , { name = "Willing"
      , tiles = Array.empty
      }
    , { name = "Maybe"
      , tiles = Array.empty
      }
    , { name = "No"
      , tiles = Array.empty
      }
    ]
    |> Array.fromList
  , mousePos = (200, 200)
  , heldTile = Just
      { hoveredColumnIndex = 0
      , hoveredTileIndex = 0
      , tile = Tile "hello!" []
      }
  }
    |> withoutCmd

subscriptions : Model -> Sub Msg
subscriptions model = Sub.batch
  [ if (isJust model.heldTile)
      then onMouseMove <|
        D.map2 NewMousePos
          (D.field "pageX" D.int)
          (D.field "pageY" D.int)
      else Sub.none
  ]

view : Model -> Browser.Document Msg
view {columns, heldTile, mousePos} =
  [ section
    [ class "columns"
    ]
    ( Array.indexedMap (viewColumn heldTile) columns
        |> Array.toList
    )
  , viewMaybe (viewFloatyTile mousePos) heldTile
  ]
    |> Browser.Document "meow!"

viewFloatyTile : (Int, Int) -> HeldTile -> Html Msg
viewFloatyTile (mouseX, mouseY) { tile, hoveredColumnIndex } =
  div
    [ style "top"  (mouseY |> String.fromInt |> flip (++) "px")
    , style "left" (mouseX |> String.fromInt |> flip (++) "px")
    , id "floaty-tile"
    , class ((++) "hovering-" <| String.fromInt <| 1 + hoveredColumnIndex)
    ]
    [ viewTile False -1 -1 tile ]


viewHoverTile = div [id "hover-tile"] []

viewColumn : Maybe HeldTile -> Int -> Column -> Html Msg
viewColumn heldOverTile columnIndex {name, tiles} =
  let
      onlyTrueActivityTiles = Array.indexedMap (viewTile (isJust heldOverTile) columnIndex) tiles
      trueActivityTilesPlusHoverTile = case heldOverTile of
        Nothing -> onlyTrueActivityTiles
        Just {hoveredColumnIndex, hoveredTileIndex} ->
          if hoveredColumnIndex == columnIndex
            then
              Array.insertAt
                hoveredTileIndex
                viewHoverTile
                onlyTrueActivityTiles
            else onlyTrueActivityTiles
      tilesList = Array.toList trueActivityTilesPlusHoverTile
  in
    div
      [ class "column"
      ]
      ( ( h3
           []
           [text name]
        ) :: tilesList
          ++ [ viewAddTile columnIndex ]
      )

viewTile : Bool -> Int -> Int -> Tile -> Html Msg
viewTile isTileHeld columnIndex tileIndex tile =
  div
    [ class "tile"
    , on "click" <| (
        D.map2 (Click columnIndex tileIndex)
          (D.field "layerX" D.int)
          (D.field "layerY" D.int)
      )
    , attributeIf isTileHeld (onMouseEnter <| HeldOverNewTile columnIndex tileIndex)
    ]
    [ input
      [ value tile.text
      , onInput (SetTileText columnIndex tileIndex)
      , onBlur (TileDeselected columnIndex)
      , attributeIf (isBlank tile.text) (id "new-tile")
      ]
      []
    , svg
      [ Svg.Attributes.class "tile-stickers"
      , height "80"
      , width "400"
      , style "user-select" "none"
      ]
      (map viewSticker tile.stickers)
    ]

viewAddTile : Int -> Html Msg
viewAddTile columnIndex =
  button
    [ class "add-tile"
    , class "tile"
    , onClick (AddTile columnIndex)
    ]
    [ text "+" ]

addStickerToTile : Sticker -> Tile -> Tile
addStickerToTile sticker tile =
  { tile | stickers = sticker :: tile.stickers }

addTileToColumn : Column -> Column
addTileToColumn column =
  { column | tiles = Array.push blankTile column.tiles }

isTileEmpty : Tile -> Bool
isTileEmpty = .text >> isEmpty

pruneTiles : Column -> Column
pruneTiles column =
  { column | tiles = Array.filter (not << isTileEmpty) column.tiles }

updateTileText : String -> Tile -> Tile
updateTileText text tile =
  { tile | text = text }

modTileInColumn : Int -> (Tile -> Tile) -> Column -> Column
modTileInColumn tileIndex f column =
  { column | tiles =
    Array.update
      tileIndex
      f
      column.tiles
  }

modColumnInPage : Int -> (Column -> Column) -> Model -> Model
modColumnInPage columnIndex f model =
  { model | columns =
    Array.update
      columnIndex
      f
      model.columns
  }

update : Msg -> Model -> (Model, Cmd Msg)
update msg model = case Debug.log "UPDATE" msg of
  Noop -> (model, Cmd.none)
  Click columnIndex tileIndex x y ->
    (modColumnInPage columnIndex << modTileInColumn tileIndex)
      (addStickerToTile <| Sticker (TextSticker "red" "<3") "Into This" x y)
      model
      |> withoutCmd
  SetTileText columnIndex tileIndex newText ->
    (modColumnInPage columnIndex << modTileInColumn tileIndex)
      (updateTileText newText)
      model
      |> withoutCmd
  AddTile columnIndex ->
    modColumnInPage columnIndex
      addTileToColumn
      model
      |> withCmd (Task.attempt (constant Noop) (focus "new-tile"))
  TileDeselected columnIndex ->
    modColumnInPage columnIndex
      pruneTiles
      model
      |> withoutCmd
  HeldOverNewTile columnIndex tileIndex -> case model.heldTile of
    Nothing -> model |> withoutCmd
    Just oldHeldTileInfo ->
      let
        newTileIndex =
          if oldHeldTileInfo.hoveredTileIndex == tileIndex
            then tileIndex + 1
            else tileIndex
      in
        { model
        | heldTile =
          Just
            { oldHeldTileInfo
            | hoveredColumnIndex = columnIndex
            , hoveredTileIndex = newTileIndex
            }
        }
          |> withoutCmd
  NewMousePos x y ->
    { model | mousePos = (x, y) }
    |> withoutCmd